package com.vaadin.spring.roo.addon.entityform;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;
import java.util.SortedMap;
import java.util.logging.Logger;

import org.apache.felix.scr.annotations.Component;
import org.apache.felix.scr.annotations.Reference;
import org.springframework.roo.addon.jpa.activerecord.JpaActiveRecordMetadata;
import org.springframework.roo.addon.web.mvc.controller.details.JavaTypeMetadataDetails;
import org.springframework.roo.addon.web.mvc.controller.details.WebMetadataService;
import org.springframework.roo.classpath.PhysicalTypeIdentifier;
import org.springframework.roo.classpath.PhysicalTypeMetadata;
import org.springframework.roo.classpath.details.ClassOrInterfaceTypeDetails;
import org.springframework.roo.classpath.itd.AbstractItdMetadataProvider;
import org.springframework.roo.classpath.itd.ItdTypeDetailsProvidingMetadataItem;
import org.springframework.roo.classpath.scanner.MemberDetails;
import org.springframework.roo.model.JavaPackage;
import org.springframework.roo.model.JavaType;
import org.springframework.roo.project.LogicalPath;

import com.vaadin.spring.roo.addon.VaadinRooUtils;
import com.vaadin.spring.roo.addon.abstractentityview.VaadinAbstractEntityViewMetadata;
import com.vaadin.spring.roo.addon.annotations.RooVaadinAbstractEntityView;

/**
 * Abstract base class for Vaadin entity form metadata providers.
 */
@Component(componentAbstract = true, immediate = true)
public abstract class AbstractVaadinEntityFormMetadataProvider extends
		AbstractItdMetadataProvider {

	@Reference
	private WebMetadataService webMetadataService;

	private static Logger logger = Logger
			.getLogger(AbstractVaadinEntityFormMetadataProvider.class.getName());

	protected ItdTypeDetailsProvidingMetadataItem constructMetadata(
			String metadataIdentificationString, JavaType aspectName,
			PhysicalTypeMetadata governorPhysicalTypeMetadata,
			String itdFilename, JavaType formBackingObject) {

		final MemberDetails formBackingObjectMemberDetails = getMemberDetails(formBackingObject);
		if (formBackingObjectMemberDetails == null) {
			return null;
		}

		// We know governor type details are non-null and can be safely cast

		// Lookup the form backing object's metadata
		LogicalPath path = typeLocationService.getTypePath(formBackingObject);

		// String physicalTypeMetadataKey = PhysicalTypeIdentifier
		// .createIdentifier(formBackingObject, path);
		String jpaActiveRecordMetadataKey = JpaActiveRecordMetadata
				.createIdentifier(formBackingObject, path);

		// We need to lookup the metadata we depend on
		JpaActiveRecordMetadata jpaActiveRecordMetadata = (JpaActiveRecordMetadata) metadataService
				.get(jpaActiveRecordMetadataKey);

		// We need to abort if we couldn't find dependent metadata
		if (jpaActiveRecordMetadata == null
				|| !jpaActiveRecordMetadata.isValid()) {
			return null;
		}

		// TODO might need also physical type to handle deletions etc?

		// We need to be informed if our dependent metadata changes
		// metadataDependencyRegistry.registerDependency(physicalTypeMetadataKey,
		// metadataIdentificationString);
		metadataDependencyRegistry.registerDependency(
				jpaActiveRecordMetadataKey, metadataIdentificationString);

		// We also need to be informed if a referenced type is changed
		MemberDetails memberDetails = getMemberDetails(formBackingObject);
		Collection<JavaType> specialDomainTypes = VaadinRooUtils
				.getSpecialDomainTypes(metadataService, formBackingObject,
						memberDetails, true).values();
		if (specialDomainTypes == null) {
			return null;
		}
		// unique types only
		for (JavaType type : new LinkedHashSet<JavaType>(specialDomainTypes)) {
			// TODO what metadata to depend on?
			metadataDependencyRegistry.registerDependency(
					PhysicalTypeIdentifier.createIdentifier(type, path),
					metadataIdentificationString);
		}

		// find any class with the annotation @RooVaadinAbstractEntityView
		// same package will contain helper classes

		// TODO refactor the scanning to be common between all such locations,
		// handle multiple hits
		VaadinAbstractEntityViewMetadata abstractViewMetadata = null;

		Set<ClassOrInterfaceTypeDetails> abstractEntityViews = typeLocationService
				.findClassesOrInterfaceDetailsWithAnnotation(new JavaType(
						RooVaadinAbstractEntityView.class));
		// must have the package and JPAContainer use information
		if (abstractEntityViews.size() == 0) {
			return null;
		}
		// pick the first one
		JavaType abstractEntityViewType = abstractEntityViews.iterator().next()
				.getType();
		String mid = VaadinAbstractEntityViewMetadata.createIdentifier(
				abstractEntityViewType,
				typeLocationService.getTypePath(abstractEntityViewType));
		if (VaadinAbstractEntityViewMetadata.isValid(mid)) {
			abstractViewMetadata = (VaadinAbstractEntityViewMetadata) metadataService
					.get(mid);
		}

		// must have the package and JPAContainer use information
		if (abstractViewMetadata == null) {
			return null;
		}

		boolean useJpaContainer = abstractViewMetadata.isUseJpaContainer();

		JavaType abstractViewClass = VaadinAbstractEntityViewMetadata
				.getJavaType(abstractViewMetadata.getId());
		JavaPackage utilPackage = abstractViewClass.getPackage();

		final SortedMap<JavaType, JavaTypeMetadataDetails> relatedApplicationTypeMetadata = getWebMetadataService()
				.getRelatedApplicationTypeMetadata(formBackingObject,
						formBackingObjectMemberDetails,
						metadataIdentificationString);
		final JavaTypeMetadataDetails javaTypeMetadataHolder = relatedApplicationTypeMetadata
				.get(formBackingObject);

		return createMetadataItem(metadataIdentificationString, aspectName,
				governorPhysicalTypeMetadata, formBackingObject,
				jpaActiveRecordMetadata, utilPackage, useJpaContainer,
				javaTypeMetadataHolder);
	}

	protected abstract ItdTypeDetailsProvidingMetadataItem createMetadataItem(
			String metadataIdentificationString, JavaType aspectName,
			PhysicalTypeMetadata governorPhysicalTypeMetadata,
			JavaType formBackingObject, JpaActiveRecordMetadata entityMetadata,
			JavaPackage utilPackage, boolean useJpaContainer,
			JavaTypeMetadataDetails javaTypeMetadataDetails);

	public WebMetadataService getWebMetadataService() {
		return webMetadataService;
	}

}
